from typing import Optional
import more_itertools

from . import reader
from . import types


def read(x: str) -> Optional[types.Value]:
    stream = more_itertools.peekable(x)
    res = reader.read(stream, eof_error_p=False, eof_value=types.ValueSymbol(value='EOF'))
    if isinstance(res, types.ValueSymbol) and res.value == 'EOF':
        return None
    return res


def eval(x: Optional[types.Value]) -> Optional[types.Value]:
    return x


def print(x: Optional[types.Value]) -> Optional[str]:
    if x is None:
        return None

    return x.lisfy()


def rep(x: str) -> Optional[str]:
    return print(eval(read(x)))



from __future__ import annotations

import typing
from collections.abc import Callable


T = typing.TypeVar('T')


def trap(fn: Callable[[], T]) -> tuple[T, None] | tuple[None, Exception]:
    try:
        return fn(), None
    except Exception as e:
        return None, e



from __future__ import annotations

from typing import Literal, Optional
import typing

import more_itertools

from .. import types


def skip_whitespace(input_stream: more_itertools.peekable[str]) -> None:
    while (peek := input_stream.peek(None)) and peek.isspace():
        next(input_stream)


def skip_whitespace_and_ensure(input_stream: more_itertools.peekable[str], expected: str) -> None:
    skip_whitespace(input_stream)
    peek = input_stream.peek(None)

    if peek is None:
        raise types.ReaderError('Unexpected EOF')

    if peek != expected:
        raise types.ReaderError(f'Expected {expected}, but got: {peek}')

    next(input_stream)


def ensure_char(
    ensure_char: str,
    input_stream: more_itertools.peekable[str],
    eof_error_p: bool = True,
    recursive_p: bool = False,
) -> None:
    peek = peek_char(True, input_stream, False, 'EOF', recursive_p=recursive_p)

    if peek == 'EOF':
        if eof_error_p:
            raise types.ReaderError('Unexpected EOF')
        return

    if peek != ensure_char:
        raise types.ReaderError(f'Expected {ensure_char}, but got: {peek}')


def peek_char(
    peek_type: None | Literal[True] | str,
    input_stream: more_itertools.peekable[str],
    eof_error_p: bool = True,
    eof_value: Optional[str] = None,
    recursive_p: bool = False,
) -> str:
    """
    PEEK_TYPE:
        None: peek one char and return it
        True: skip whitespace and peek one char and return it
        str: skip until PEEK_TYPE and peek one char and return it

    EOF_ERROR_P:
        True: raise an error when EOF is reached
        False: return EOF_VALUE when EOF is reached
               EOF_VALUE should be specified when EOF_ERROR_P is False
    """
    while True:
        peek = input_stream.peek(None)
        if peek is None:
            break

        if peek_type is None:
            break

        elif peek_type == True:
            if not peek.isspace():
                break

        elif isinstance(peek_type, str):  # pyright: ignore [reportUnnecessaryIsInstance]
            if peek == peek_type:
                break

        else:
            typing.assert_never(peek_type)

        next(input_stream)

    if peek is None:
        if eof_error_p:
            raise types.ReaderError('Unexpected EOF')

        if eof_value is None:
            raise ValueError('eof_value must be specified if eof_error_p is False')

        return eof_value

    return peek



from __future__ import annotations

import typing
from typing import Iterable
from collections.abc import Callable

import more_itertools


T = typing.TypeVar('T')


def takewhile_inclusive(pred: Callable[[T], bool], peekable: more_itertools.peekable[T]) -> Iterable[T]:
    while True:
        try:
            x = peekable.peek()
        except StopIteration:
            break

        if pred(x):
            yield next(peekable)
        else:
            break



from __future__ import annotations
from typing import Optional

import more_itertools

from . import types
from . import subr


def read_string(input_stream: more_itertools.peekable[str]) -> types.ValueString:
    next(input_stream)  # Skip the opening '"'.
    value = ''

    while True:
        peek = subr.reader.peek_char(None, input_stream, False, 'EOF', recursive_p=True)

        if peek == 'EOF':
            raise types.ReaderError('Unexpected EOF')

        if peek == '"':
            break

        value += peek
        next(input_stream)

    next(input_stream)  # Skip the closing '"'.
    return types.ValueString(value=value)


def read_number(input_stream: more_itertools.peekable[str]) -> types.ValueInteger | types.ValueFloat:
    value = ''

    while (peek := input_stream.peek(None)) and peek in '0123456789-+.eE':
        value += peek
        next(input_stream)

    i, _ = subr.subr.trap(lambda: int(value))
    if i is not None:
        return types.ValueInteger(value=i)

    f, _ = subr.subr.trap(lambda: float(value))
    if f is not None:
        return types.ValueFloat(value=f)

    raise types.ReaderError(f'Could not parse as a number: {value}')


def read_symbol(input_stream: more_itertools.peekable[str]) -> types.ValueSymbol:
    peek = subr.reader.peek_char(True, input_stream, recursive_p=True)

    if peek.lower() == 't':
        subr.reader.ensure_char('t', input_stream, recursive_p=True)
        subr.reader.ensure_char('r', input_stream, recursive_p=True)
        subr.reader.ensure_char('u', input_stream, recursive_p=True)
        subr.reader.ensure_char('e', input_stream, recursive_p=True)
        return types.ValueSymbol(value='true')

    if peek.lower() == 'f':
        subr.reader.ensure_char('f', input_stream, recursive_p=True)
        subr.reader.ensure_char('a', input_stream, recursive_p=True)
        subr.reader.ensure_char('l', input_stream, recursive_p=True)
        subr.reader.ensure_char('s', input_stream, recursive_p=True)
        subr.reader.ensure_char('e', input_stream, recursive_p=True)
        return types.ValueSymbol(value='false')

    if peek.lower() == 'n':
        subr.reader.ensure_char('n', input_stream, recursive_p=True)
        subr.reader.ensure_char('u', input_stream, recursive_p=True)
        subr.reader.ensure_char('l', input_stream, recursive_p=True)
        subr.reader.ensure_char('l', input_stream, recursive_p=True)
        return types.ValueSymbol(value='null')

    raise types.ReaderError(f'Unexpected char: {peek}')


def read_object(input_stream: more_itertools.peekable[str]) -> types.ValueObject:
    next(input_stream)  # Skip the opening '{'.

    peek = subr.reader.peek_char(True, input_stream, recursive_p=True)

    if peek == '}':
        next(input_stream)  # Skip the closing '}'.
        return types.ValueObject(value={})

    value = {}

    while True:
        key = read(input_stream, recursive_p=True)

        subr.reader.ensure_char(':', input_stream)
        value[key] = read(input_stream, recursive_p=True)

        peek = subr.reader.peek_char(True, input_stream, recursive_p=True)

        if peek == '}':
            break

        if peek != ',':
            raise types.ReaderError(f'Expected a comma or closing brace, but got: {peek}')

        next(input_stream)  # Skip the comma.

    next(input_stream)  # Skip the closing '}'.
    return types.ValueObject(value=value)


def read_array(input_stream: more_itertools.peekable[str]) -> types.ValueArray:
    next(input_stream)  # Skip the opening '['.

    peek = subr.reader.peek_char(True, input_stream, recursive_p=True)

    if peek == ']':
        next(input_stream)  # Skip the closing ']'.
        return types.ValueArray(value=[])

    value: list[types.Value] = []

    while True:
        value.append(read(input_stream, recursive_p=True))

        peek = subr.reader.peek_char(True, input_stream, recursive_p=True)

        if peek == ']':
            break

        if peek != ',':
            raise types.ReaderError(f'Expected a comma or closing bracket, but got: {peek}')

        next(input_stream)  # Skip the comma.

    next(input_stream)  # Skip the closing ']'.
    return types.ValueArray(value=value)


def read(
    input_stream: more_itertools.peekable[str],
    eof_error_p: bool = True,
    eof_value: Optional[types.Value] = None,
    recursive_p: bool = False,
) -> types.Value:
    peek = subr.reader.peek_char(True, input_stream, False, 'EOF', recursive_p=recursive_p)

    if peek == 'EOF':
        if eof_error_p:
            raise types.ReaderError('Unexpected EOF')

        if eof_value is None:
            raise ValueError('eof_value must be provided if eof_error_p is False')

        return eof_value

    if peek == '{':
        return read_object(input_stream)

    if peek == '}':
        raise types.ReaderError('Unexpected }')

    if peek == '[':
        return read_array(input_stream)

    if peek == ']':
        raise types.ReaderError('Unexpected ]')

    if peek == '"':
        return read_string(input_stream)

    if peek in '0123456789-+':
        return read_number(input_stream)

    return read_symbol(input_stream)



from typing import Literal
import pydantic


## Exceptions

class LisfyError(Exception):
    pass


class ReaderError(LisfyError):
    pass


## Values

class Value(pydantic.BaseModel):
    def lisfy(self, minify: bool=False) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        return str(self.__dict__).__hash__()


class ValueObject(Value):
    value: dict[Value, Value]

    def lisfy(self, minify: bool = False) -> str:
        res = [
            (k.lisfy(minify=minify), v.lisfy(minify=minify))
            for k, v
            in self.value.items()
        ]
        return '(object nil ' + ' '.join(f'(item nil {k} {v})' for k, v in res) + ')'


class ValueArray(Value):
    value: list[Value]

    def lisfy(self, minify: bool=False) -> str:
        res = [x.lisfy(minify=minify) for x in self.value]
        return '(array nil ' + ' '.join(res) + ')'


class ValueString(Value):
    value: str

    def lisfy(self, minify: bool=False) -> str:
        return f'(str nil "{self.value}")'


class ValueInteger(Value):
    value: int

    def lisfy(self, minify: bool=False) -> str:
        return f'(int nil {str(self.value)})'


class ValueFloat(Value):
    value: float

    def lisfy(self, minify: bool=False) -> str:
        return f'(float nil {str(self.value)})'


class ValueSymbol(Value):
    value: Literal['EOF', 'null', 'true', 'false']

    def lisfy(self, minify: bool=False) -> str:
        if self.value == 'EOF':
            raise LisfyError('Cannot lisfy EOF')

        return f'(symbol nil "{self.value}")'



from . import rep

def main():
    while True:
        try:
            text = input('json-lisfy> ')
        except (EOFError, KeyboardInterrupt):
            break
        if (val := rep.rep(text)):
            print(val)
